from __future__ import annotations

import os
import shutil
import urllib.parse, urllib.error
from dataclasses import dataclass
from functools import cached_property
from io import StringIO
from pathlib import Path
from typing import Union, Literal, Optional, Type

import pandas as pd
import yaml
import requests

from hdxms_datasets.process import filter_peptides
from hdxms_datasets.reader import read_dynamx

from hdxms_datasets.config import cfg


@dataclass(frozen=True)
class DataFile(object):

    name: str

    format: Literal["DynamX"]

    filepath_or_buffer: Union[Path, StringIO]

    @cached_property
    def data(self) -> pd.DataFrame:
        if self.format == "DynamX":
            # from, to time conversion
            time_conversion = (cfg.dynamx.time_unit, cfg.time_unit)

            data = read_dynamx(self.filepath_or_buffer, time_conversion=time_conversion)
        else:
            raise ValueError(f"Invalid format {self.format!r}")

        if isinstance(self.filepath_or_buffer, StringIO):
            self.filepath_or_buffer.seek(0)

        return data


class StateParser(object):
    """

    Args:
        hdx_spec: Dictionary with HDX-MS state specification.
        data_src: Optional data source with input data files. If not specified, current
            directory is used. Otherwise, either a data source path can be specified or
            data can be given as a dictionary, where keys are filenames and values are
            :class:`~io.StringIO` with file contents.
    """

    def __init__(
        self,
        hdx_spec: dict,
        data_src: Union[os.PathLike[str], str, dict[str, DataFile], None],
    ) -> None:

        self.hdx_spec = hdx_spec
        self.data_files: dict[str, DataFile] = {}

        if isinstance(data_src, (os.PathLike, str)):
            data_src = Path(data_src) or Path(".")
            for name, spec in self.hdx_spec["data_files"].items():
                datafile = DataFile(
                    name=name,
                    filepath_or_buffer=data_src / spec["filename"],
                    **{k: v for k, v in spec.items() if k != "filename"},
                )
                self.data_files[name] = datafile

        elif isinstance(data_src, dict):
            self.data_files = data_src
        else:
            raise TypeError(f"Invalid data type {type(data_src)!r}, must be path or dict")

    def load_peptides(self, state: Union[str, int], peptides: str) -> pd.DataFrame:
        # todo allow peptides as int, None
        state = self.states[state] if isinstance(state, int) else state
        peptide_spec = self.hdx_spec["states"][state]["peptides"][peptides]

        df = self.data_files[peptide_spec["data_file"]].data

        filter_fields = {"state", "exposure", "query", "dropna"}
        peptides = filter_peptides(
            df, **{k: v for k, v in peptide_spec.items() if k in filter_fields}
        )

        return peptides

    @property
    def states(self) -> list[str]:
        return list(self.hdx_spec["states"].keys())

    @property
    def state_peptide_sets(self) -> dict[str, list[str]]:
        """Dictionary of state names and list of peptide sets for each state"""
        return {state: list(spec["peptides"]) for state, spec in self.hdx_spec["states"].items()}

    @cached_property
    def peptide_sets(self) -> dict[str, dict[str, pd.DataFrame]]:

        peptides_dfs = {}
        for state, peptides in self.state_peptide_sets.items():
            peptides_dfs[state] = {
                peptide_set: self.load_peptides(state, peptide_set) for peptide_set in peptides
            }

        return peptides_dfs


# TODO remove parser and merge with HDXDataSet
@dataclass
class HDXDataSet(object):

    parser: StateParser

    data_id: str

    metadata: Optional[dict]

    def describe(self):
        """
        Returns states and peptides in the dataset
        """
        ...
        output = {}

    def cite(self) -> str:
        """
        Returns citation information
        """

        return "Not implemented"

    def __len__(self) -> int:
        return len(self.parser.states)


class DataVault(object):
    def __init__(
        self,
        cache_dir: Optional[Union[Path[str], str]] = None,
        parser: Type[StateParser] = StateParser,
    ):

        if cache_dir is None:
            self.cache_dir = Path.home() / ".hdxms_datasets" / "datasets"
            self.cache_dir.mkdir(exist_ok=True, parents=True)
        else:
            self.cache_dir: Path = Path(cache_dir)
            if not self.cache_dir.exists():
                raise FileNotFoundError(f"Cache directory '{self.cache_dir}' does not exist")

        self.parser = parser

    def filter(self, *spec: dict):
        # filters list of available datasets
        ...

    @cached_property
    def index(self) -> list[str]:
        """List of available datasets in the remote database"""

        url = urllib.parse.urljoin(cfg.database_url, "index.txt")
        response = requests.get(url)
        if response.ok:
            index = response.text.split("\n")[1:]
            return index
        else:
            return []

    @property
    def datasets(self) -> list[str]:
        """List of available datasets in the cache dir"""
        return [d.stem for d in self.cache_dir.iterdir() if self.is_dataset(d)]

    @staticmethod
    def is_dataset(path: Path) -> bool:
        """
        Checks if the supplied path is a HDX-MS dataset.
        """

        return (path / "hdx_spec.yaml").exists()

    async def fetch_datasets(self, n: Optional[str] = None, data_ids: Optional[list[str]] = None):
        """
        Asynchronously download multiple datasets
        """
        if n is None and data_ids is None:
            n = 10

        if data_ids:
            # Download specified datasets to cache_dir
            ...

        elif n:
            # Download n new datasets to cache_dir
            ...

    def fetch_dataset(self, data_id: str) -> bool:
        """
        Download a dataset from the online repository to the cache dir

        :param data_id:
        :return:
        """

        output_pth = cfg.database_dir / data_id
        if output_pth.exists():
            return False
        else:
            output_pth.mkdir()

        dataset_url = urllib.parse.urljoin(cfg.database_url, data_id + "/")

        files = ["hdx_spec.yaml", "metadata.yaml"]
        optional_files = ["CITATION.cff"]
        for f in files + optional_files:
            url = urllib.parse.urljoin(dataset_url, f)
            response = requests.get(url)

            if response.ok:
                (output_pth / f).write_bytes(response.content)

            elif f in files:
                raise urllib.error.HTTPError(
                    url, response.status_code, f"Error for file {f!r}", response.headers, None
                )

            if f == "hdx_spec.yaml":
                hdx_spec = yaml.safe_load(response.text)

        data_pth = output_pth / "data"
        data_pth.mkdir()

        for file_spec in hdx_spec["data_files"].values():
            filename = file_spec["filename"]
            f_url = urllib.parse.urljoin(dataset_url, filename)
            response = requests.get(f_url)

            if response.ok:
                (output_pth / filename).write_bytes(response.content)
            else:
                raise urllib.error.HTTPError(
                    f_url,
                    response.status_code,
                    f"Error for data file {filename!r}",
                    response.headers,
                    None,
                )

        return True

    def clear_cache(self) -> None:
        for dir in self.cache_dir.iterdir():
            shutil.rmtree(dir)

    def load_dataset(self, data_id: str) -> HDXDataSet:
        ...
        # if not in cache dir -> fetch

        hdx_spec = yaml.safe_load((self.cache_dir / data_id / "hdx_spec.yaml").read_text())
        parser = self.parser(hdx_spec, self.cache_dir / data_id)
        metadata = yaml.safe_load((self.cache_dir / data_id / "metadata.yaml").read_text())
        return HDXDataSet(parser=parser, data_id=data_id, metadata=metadata)



"""Top-level package for HDXMS Datasets."""

from hdxms_datasets.datasets import DataVault, StateParser
from hdxms_datasets.config import cfg

# placeholder version number
__version__ = "0.1.0a1"

# when we are on editable install from source, the _version file is present
# and we can get a version from there
try:
    from . import _version

    __version__ = _version.get_versions()["version"]
except ImportError:
    pass



from __future__ import annotations

import os
from pathlib import Path
from typing import Any

from omegaconf import OmegaConf, DictConfig, DictKeyType
from packaging import version
from hdxms_datasets._version import get_versions

__version__ = get_versions()["version"]
del get_versions


def reset_config():
    """Create a new config.yaml file in the user home dir/.hdxms_datasets folder"""

    with open(conf_home_pth, "w") as target:
        version_string = "# HDXMS datasets configuration file " + __version__ + "\n\n"
        target.write(version_string)

        with open(current_dir / "config.yaml") as source:
            for line in source:
                target.write(line)


class Singleton(type):
    _instances: dict[type, Singleton] = {}

    def __call__(cls, *args: Any, **kwargs: Any) -> Any:
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]

    def instance(cls: Any, *args: Any, **kwargs: Any) -> Any:
        return cls(*args, **kwargs)


class HDXMSDatasetsConfig(metaclass=Singleton):
    __slots__ = ["conf"]

    def __init__(self) -> None:
        self.conf = None

    def __getattr__(self, item: str) -> Any:
        return getattr(self.conf, item)

    def load_config(self, config_file: os.PathLike[str]):
        conf = OmegaConf.create(Path(config_file).read_text())
        self.set_config(conf)

    def set_config(self, conf: DictConfig) -> None:
        self.conf = conf

    def get(self, key: DictKeyType, default_value: Any = None) -> Any:
        return self.conf.get(key, default_value)

    @property
    def database_dir(self) -> Path:
        pth = self.conf.database_dir
        if "~" in pth:
            database_dir = Path(pth.replace("~", str(Path.home())))
        elif "$home" in pth:
            database_dir = Path(pth.replace("$home", str(Path.home())))
        else:
            database_dir = Path(pth)

        return database_dir


def valid_config() -> bool:
    """Checks if the current config file in the user home directory is a valid config
    file for the current hdxms_datasets version

    """
    if not conf_home_pth.exists():
        return False
    else:
        with open(conf_home_pth, "r") as f:
            version_string = f.readline().strip("; ").split(" ")[-1]

        hdxms_datasets_version = version.parse(__version__)
        cfg_version = version.parse(version_string)

        return hdxms_datasets_version.public == cfg_version.public


# https://stackoverflow.com/questions/6198372/most-pythonic-way-to-provide-global-configuration-variables-in-config-py/25880082
class CfgClass(metaclass=Singleton):
    def __init__(self, config=None):
        self._config = {} if config is None else config

    def __getitem__(self, item):
        return self._config[item]


home_dir = Path.home()
config_dir = home_dir / ".hdxms_datasets"
config_dir.mkdir(parents=False, exist_ok=True)
conf_home_pth = config_dir / "config.yaml"

current_dir = Path(__file__).parent
conf_src_pth = current_dir / "config.yaml"

# Current config version is outdated
if not valid_config():
    try:
        reset_config()
        conf = OmegaConf.load(conf_home_pth)
    except FileNotFoundError:
        # This will happen on conda-forge docker build.
        # When no config.yaml file is in home_dir / '.dont_fret',
        # ConfigurationSettings will use the hardcoded version
        conf = OmegaConf.load(conf_src_pth)
        # (this is run twice due to import but should be OK since conf is singleton)
else:
    conf = OmegaConf.load(conf_home_pth)


cfg = HDXMSDatasetsConfig()
cfg.set_config(conf)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['hdxms_datasets']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'omegaconf>=2.3.0,<3.0.0',
 'packaging>=22.0,<23.0',
 'pandas>=1.5.2,<2.0.0',
 'requests>=2.28.1,<3.0.0']

setup_kwargs = {
    'name': 'hdxms-datasets',
    'version': '0.1.0a1',
    'description': 'Download and parse curated HDX-MS datasets',
    'long_description': '# HDXMS Datasets\n\n\n* Free software: MIT license\n\n\n\nExample code\n--------\n\n```python\nfrom hdxms_datasets import DataVault\n\nvault = DataVault()\n\n# Download a specific HDX dataset\nvault.fetch_dataset("20221007_1530_SecA_Krishnamurthy")\n\n# Load the dataset\nds = vault.load_dataset("20221007_1530_SecA_Krishnamurthy")\n\n# Load the FD control of the first \'state\' in the dataset.\nfd_control = ds.parser.load_peptides(0, \'FD_control\')\n\n# Load the corresponding experimental peptides.\npeptides = ds.parser.load_peptides(0, \'experiment\')\n\n```',
    'author': 'Jochem Smit',
    'author_email': 'jhsmit@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



"""
Reader functions for various file formats
"""

from __future__ import annotations

from io import StringIO
from pathlib import Path
from typing import Union, Literal
import pandas as pd


def read_dynamx(
    filepath_or_buffer: Union[Path[str], str, StringIO],
    time_conversion: tuple[Literal["h", "min", "s"], Literal["h", "min", "s"]] = ("min", "s"),
) -> pd.DataFrame:

    """
    Reads DynamX .csv files and returns the resulting peptide table as a pandas DataFrame.

    Args:
        filepath_or_buffer: File path of the .csv file or :class:`~io.StringIO` object.
        time_conversion: How to convert the time unit of the field 'exposure'. Format is ('<from>', <'to'>).
            Unit options are 'h', 'min' or 's'.

    Returns:
        Peptide table as a pandas DataFrame.
    """

    if isinstance(filepath_or_buffer, StringIO):
        hdr = filepath_or_buffer.readline().strip("# \n\t")
        filepath_or_buffer.seek(0)
    else:
        with open(filepath_or_buffer, "r") as f_obj:
            hdr = f_obj.readline().strip("# \n\t")

    names = [name.lower().strip("\r\t\n") for name in hdr.split(",")]
    df = pd.read_csv(filepath_or_buffer, header=0, names=names)

    df.insert(df.columns.get_loc("end") + 1, "stop", df["end"] + 1)

    time_lut = {"h": 3600, "min": 60, "s": 1}
    time_factor = time_lut[time_conversion[0]] / time_lut[time_conversion[1]]

    df["exposure"] *= time_factor
    df.columns = df.columns.str.replace(" ", "_")

    return df



from __future__ import annotations

import pandas as pd
import numpy.typing as npt
from hdxms_datasets.config import cfg

from typing import Literal, Optional, Union, TypeVar


time_factors = {"s": 1, "m": 60.0, "min": 60.0, "h": 3600, "d": 86400}
temperature_offsets = {"c": 273.15, "celsius": 273.15, "k": 0.0, "kelvin": 0.0}

A = TypeVar("A", npt.ArrayLike, pd.Series, pd.DataFrame)


def convert_time(
    values: A, src_unit: Literal["h", "min", "s"], target_unit: Literal["h", "min", "s"]
) -> A:

    time_lut = {"h": 3600.0, "min": 60.0, "s": 1.0}
    time_factor = time_lut[src_unit] / time_lut[target_unit]

    if isinstance(values, list):
        return [v * time_factor for v in values]
    else:
        return values * time_factor


def filter_peptides(
    df: pd.DataFrame,
    state: Optional[str] = None,
    exposure: Union[dict, float, None] = None,
    query: Optional[list[str]] = None,
    dropna: bool = True,
) -> pd.DataFrame:
    """
    Convenience function to filter a peptides DataFrame.

    Args:
        df: Input :class:`pandas.DataFrame`
        state: Name of protein state to select.
        exposure: Exposure value(s) to select. Exposure is given as a :obj:`dict`, with keys "value" or "values" for
            exposure value, and "unit" for the time unit.
        query: Additional queries to pass to :meth:`pandas.DataFrame.query`.
        dropna: Drop rows with NaN uptake entries.

    Example:
        ::

        d = {"state", "SecB WT apo", "exposure": {"value": 0.167, "unit": "min"}
        filtered_df = filter_peptides(df, **d)

    Returns:

    """

    if state:
        df = df[df["state"] == state]

    if isinstance(exposure, dict):
        if values := exposure.get("values"):
            values = convert_time(values, exposure.get("unit", "s"), cfg.time_unit)
            df = df[df["exposure"].isin(values)]
        elif value := exposure.get("value"):
            value = convert_time(value, exposure.get("unit", "s"), cfg.time_unit)
            df = df[df["exposure"] == value]
    elif isinstance(exposure, float):
        df = df[df["exposure"] == exposure]

    if query:
        for q in query:
            df = df.query(q)

    if dropna:
        df = df.dropna(subset=["uptake"])

    return df
